import json
import sys
import time
import threading
import subprocess
from typing import cast
import click
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler
import os
import rich
import multiprocessing
from relationalai import metagen
import tempfile
import webbrowser
import shutil

from . import snapshot_viewer as viewer
from . import cleanup_snapshots as cleanup


from .cli_controls import divider, Spinner

#--------------------------------------------------
# Root
#--------------------------------------------------

@click.group()
def cli():
    pass

#--------------------------------------------------
# Watch
#--------------------------------------------------

def clear():
    os.system('cls' if os.name == 'nt' else 'clear')

class ChangeHandler(PatternMatchingEventHandler):

    def __init__(self):
        super().__init__(patterns = ["*.py", "*.rel", "raiconfig.toml"])
        self.script = None
        self.process = None
        self.event_lock = threading.Lock()
        self.has_queued_events = False

    def check_event(self, event):
        if not event.src_path.endswith(".py"):
            return

        if "examples/" in event.src_path \
            or "tests/end2end/test_cases/" in event.src_path \
            or "tests/snowflake_integration/test_cases/" in event.src_path \
            or "metagen.py" in event.src_path:
            self.script = event.src_path

    def on_any_event(self, event):
        self.check_event(event)
        with self.event_lock:
            if self.process and self.process.poll() is None:
                # Mark that there are queued events
                self.has_queued_events = True
            else:
                self.start_process()

    def start_process(self):
        if self.script is None:
            return

        clear()
        rich.print(f"[yellow bold]{os.path.basename(self.script)}")
        rich.print("[yellow]------------------------------------------------------")
        rich.print("")

        # Start or restart the script
        self.process = subprocess.Popen(['python', self.script], shell=False)
        # Use a thread to wait for the process to finish without blocking
        wait_thread = threading.Thread(target=self.wait_and_restart_if_needed)
        wait_thread.start()

    def wait_and_restart_if_needed(self):
        if self.process is not None:
            self.process.wait()

        with self.event_lock:
            if self.has_queued_events:
                # Reset the flag and restart the process for batched events
                self.has_queued_events = False
                # Delay added to allow for potentially more events to accumulate
                # Adjust or remove the delay as needed
                time.sleep(0.5)
                self.start_process()

@cli.command()
@click.argument('directory', type=click.Path(exists=True))
def watch(directory):
    """Watch a DIRECTORY and re-run a SCRIPT on file changes."""
    event_handler = ChangeHandler()
    observer = Observer()
    observer.schedule(event_handler, path=directory, recursive=True)  # Now recursive
    observer.start()

    clear()
    rich.print(f"[yellow]Watching for changes in '{directory}'.")
    rich.print("[yellow]Save a script in examples/ to run it.")
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()

    observer.join()

#--------------------------------------------------
# Code stats
#--------------------------------------------------

def cloc(paths):
    script_dir = os.path.dirname(os.path.realpath(__file__))
    src_dir = os.path.join(script_dir, '..')
    core_paths = [os.path.join(src_dir, f) for f in paths]
    process = subprocess.Popen(['cloc', '--json', *core_paths], stdout=subprocess.PIPE, text=True)
    output, _ = process.communicate()

    try:
        # Parse the JSON output
        data = json.loads(output)
        # The total number of source lines is under 'SUM'/'code'
        total_lines = data['SUM']['code'] if 'SUM' in data else 0
        return cast(int, total_lines)
    except Exception as e:
        print(f"Error while parsing cloc output: {e}")
        return 0

@cli.command()
def stats():
    dsl = cloc(["dsl.py"])
    compiler = cloc(["compiler.py"])
    metamodel = cloc(["metamodel.py"])
    emitter = cloc(["rel_emitter.py"])
    rel = cloc(["rel.py"])
    metagen = cloc(["metagen.py"])
    gentest = cloc(["metagen.py"])
    tools = cloc(["tools/"])
    clients = cloc(["clients/"])
    std = cloc(["std/"])
    non_test_total = cloc(["."]) - metagen
    total = non_test_total + gentest
    core = dsl + compiler + metamodel + emitter + rel

    max_width = len(f"{total:,}")

    # Print statements with numbers right-aligned
    divider()
    rich.print(f"[yellow]RelationalAI  {non_test_total:>{max_width},} loc")
    rich.print(f"[yellow]  Core        {core:>{max_width},} loc")
    rich.print(f"[yellow]    dsl       {dsl:>{max_width},} loc")
    rich.print(f"[yellow]    rel       {rel:>{max_width},} loc")
    rich.print(f"[yellow]    emitter   {emitter:>{max_width},} loc")
    rich.print(f"[yellow]    metamodel {metamodel:>{max_width},} loc")
    rich.print(f"[yellow]    compiler  {compiler:>{max_width},} loc")
    rich.print(f"[yellow]  Clients     {clients:>{max_width},} loc")
    rich.print(f"[yellow]  Std         {std:>{max_width},} loc")
    rich.print(f"[yellow]  Tools       {tools:>{max_width},} loc")
    rich.print("")
    rich.print(f"[cyan]Gentest       {gentest:>{max_width},} loc")
    rich.print("")
    rich.print(f"[magenta]All           {total:>{max_width},} loc")
    divider()


#--------------------------------------------------
# Metagen
#--------------------------------------------------

@cli.command("gen")
@click.option('--total', default=50000, help='Total number of models to generate.')
@click.option('--threads', default=multiprocessing.cpu_count(), help='Threads to use, default is CPU count.')
@click.option('--internal', default=False, is_flag=True)
def gen(total, threads, internal):
    if not internal:
        divider()
    with Spinner(f"Testing {total:,.0f} models on {threads:,.0f} threads", f"Tested {total:,.0f} models"):
        if threads > 1:
            (elapsed, results) = metagen.batches(total, threads)
        else:
            gen = metagen.batch(total)
            results = [gen]
            elapsed = gen.elapsed

    rich.print("")

    for result in results:
        rich.print(result)

    failed = False
    for result in results:
        if len(result.failures) > 0:
            failed = True
            rich.print()
            result.print_failures(1)
            rich.print("")
            break

    rich.print("")
    rich.print(f"[yellow bold]Total time: {elapsed:,.3f}s")
    rich.print("")
    if not internal:
        divider()
        sys.exit(failed)

#--------------------------------------------------
# Metagen Watch
#--------------------------------------------------

class MetagenWatcher(PatternMatchingEventHandler):

    def __init__(self, total, threads):
        super().__init__(patterns = ["*.py"])
        self.process = None
        self.event_lock = threading.Lock()
        self.has_queued_events = False
        self.total = total
        self.threads = threads
        self.start_process()

    def on_any_event(self, event):
        with self.event_lock:
            if self.process and self.process.poll() is None:
                # Mark that there are queued events
                self.has_queued_events = True
            else:
                self.start_process()

    def start_process(self):
        clear()
        rich.print("[yellow bold]Metagen")
        rich.print("[yellow]------------------------------------------------------")
        rich.print("")

        # Start or restart the script
        self.process = subprocess.Popen(['rai-dev', 'gen', '--total', str(self.total), '--threads', str(self.threads), "--internal"], shell=False)
        # Use a thread to wait for the process to finish without blocking
        wait_thread = threading.Thread(target=self.wait_and_restart_if_needed)
        wait_thread.start()

    def wait_and_restart_if_needed(self):
        if self.process is not None:
            self.process.wait()

        with self.event_lock:
            if self.has_queued_events:
                # Reset the flag and restart the process for batched events
                self.has_queued_events = False
                # Delay added to allow for potentially more events to accumulate
                # Adjust or remove the delay as needed
                time.sleep(0.5)
                self.start_process()


@cli.command("gen:watch")
@click.argument('directory', type=click.Path(exists=True))
@click.option('--total', default=20000, help='Total number of models to generate')
@click.option('--threads', default=multiprocessing.cpu_count(), help='Threads to use')
def gen_watch(directory, total, threads):
    clear()
    event_handler = MetagenWatcher(total, threads)
    observer = Observer()
    observer.schedule(event_handler, path=directory, recursive=True)  # Now recursive
    observer.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()

    observer.join()

#--------------------------------------------------
# Snapshot Management
#--------------------------------------------------

def get_modified_files():
    """Get a list of files that are modified according to git."""
    try:
        # Get both staged and unstaged changes
        staged = subprocess.check_output(['git', 'diff', '--name-only', '--cached']).decode('utf-8').splitlines()
        unstaged = subprocess.check_output(['git', 'diff', '--name-only']).decode('utf-8').splitlines()
        untracked = subprocess.check_output(['git', 'ls-files', '--others', '--exclude-standard']).decode('utf-8').splitlines()

        # Combine all changes
        return set(staged + unstaged + untracked)
    except subprocess.CalledProcessError:
        print("Warning: Failed to get modified files from git. Showing all files.")
        return None

@cli.group()
def snapshots():
    """Manage snapshot tests."""
    pass

def get_repo_root():
    """Get the root directory of the repository."""
    repo_root = os.getcwd()
    if not os.path.exists(os.path.join(repo_root, ".git")):
        print("Error: Please run this command from the repository root.")
        sys.exit(1)
    return repo_root

@snapshots.command("view")
@click.option('--test-dir', default='tests/early_access/unit', help='Test directory to scan')
@click.option('--output', default=None, help='Output HTML file (defaults to a temp file)')
@click.option('--all', is_flag=True, help='Show all snapshots, not just modified ones')
def snapshots_view(test_dir, output, all):
    """Generate an HTML report of snapshot tests and open it in a browser."""
    repo_root = get_repo_root()

    # Make sure the test_dir is an absolute path
    if not os.path.isabs(test_dir):
        test_dir = os.path.join(repo_root, test_dir)

    # Check if the test directory exists
    if not os.path.exists(test_dir):
        print(f"Error: Test directory {test_dir} does not exist")
        return

    # Create a temporary file for the HTML report if not specified
    if not output:
        temp_dir = tempfile.mkdtemp()
        output = os.path.join(temp_dir, "snapshot_report.html")
    elif not os.path.isabs(output):
        output = os.path.join(repo_root, output)

    # Find all snapshot files
    snapshot_files = []
    for root, _, files in os.walk(test_dir):
        for file in files:
            if file.endswith('.snapshot'):
                snapshot_files.append(os.path.join(root, file))

    # Filter to only show modified snapshots if needed
    if not all:
        modified_files = get_modified_files()
        if modified_files:
            # Filter to only include modified snapshot files
            modified_snapshots = []
            for path in snapshot_files:
                if any(path.endswith(f) for f in modified_files if f.endswith('.snapshot')):
                    modified_snapshots.append(path)

            if modified_snapshots:
                snapshot_files = modified_snapshots
                print(f"Filtered to {len(snapshot_files)} modified snapshot files")
            else:
                print("No modified snapshot files found. Use --all to see all snapshots.")

    # Use the registry to get test information
    registry = viewer.load_snapshot_registry()
    if not registry:
        print("No snapshot registry found. Run tests first to create it.")
        return

    # Generate the report using the registry
    viewer.generate_report_from_registry(registry, snapshot_files, output)

    print(f"Report generated: {output}")

    # Open the report in the default browser
    webbrowser.open(f"file://{os.path.abspath(output)}")

@snapshots.command("cleanup")
@click.option('--test-dir', default='tests/early_access/unit', help='Test directory to scan')
@click.option('--remove', is_flag=True, help='Remove unused snapshot files')
@click.option('--regenerate', is_flag=True, help='Regenerate snapshots by running tests')
def snapshots_cleanup(test_dir, remove, regenerate):
    """Identify and optionally remove unused snapshot files."""
    repo_root = get_repo_root()

    # Make sure the test_dir is an absolute path
    if not os.path.isabs(test_dir):
        test_dir = os.path.join(repo_root, test_dir)

    if remove and regenerate:
        # Remove and regenerate
        removed = cleanup.remove_snapshots(test_dir)
        regenerated = cleanup.regenerate_snapshots(test_dir)

        print(f"\nSummary: Removed {removed} snapshots, regenerated {regenerated} snapshots")
        if removed > regenerated:
            print(f"Cleaned up {removed - regenerated} unused snapshots")
    elif remove:
        # Just remove
        cleanup.remove_snapshots(test_dir)
    elif regenerate:
        # Just regenerate
        cleanup.regenerate_snapshots(test_dir)
    else:
        # Just count and report
        snapshots = cleanup.find_all_snapshot_files(test_dir)
        print(f"Found {len(snapshots)} snapshot files")
        print("\nUse --remove to remove all snapshots")
        print("Use --regenerate to regenerate snapshots by running tests")

#--------------------------------------------------
# Test Runner
#--------------------------------------------------

@cli.group("test")
def test_group():
    """Run tests with optional snapshot updates."""
    pass

@test_group.command("unit")
@click.argument('test_path', required=False)
@click.option('--update', is_flag=True, help='Update snapshot files')
@click.option('--verbose', '-v', is_flag=True, help='Verbose output')
def run_unit_tests(test_path, update, verbose):
    """Run unit tests with optional snapshot updates."""
    repo_root = get_repo_root()

    # Determine the test path
    if not test_path:
        test_path = "tests/unit"
    elif test_path == "early_access":
        test_path = "tests/early_access/unit"
    elif test_path == "builder":
        test_path = "tests/early_access/unit/builder"
    elif test_path == "ir":
        test_path = "tests/early_access/unit/ir"
    elif test_path == "rel":
        test_path = "tests/early_access/unit/rel"
    elif test_path == "devmode":
        test_path = "tests/early_access/unit/devmode"

    # Make sure the test_path is an absolute path
    if not os.path.isabs(test_path):
        test_path = os.path.join(repo_root, test_path)

    # Build the command
    cmd = ["pytest", test_path]

    # Add options
    if update:
        cmd.append("--snapshot-update")
    if verbose:
        cmd.append("-v")

    divider()
    rich.print("[yellow bold]Running unit tests")
    rich.print(f"[yellow]Test path: {test_path}")
    rich.print("")

    # Run the command
    result = subprocess.run(cmd)

    divider()
    sys.exit(result.returncode)

@test_group.command("e2e")
@click.argument('test_path', required=False)
@click.option('--update', is_flag=True, help='Update snapshot files')
@click.option('--verbose', '-v', is_flag=True, help='Verbose output')
def run_e2e_tests(test_path, update, verbose):
    """Run end-to-end tests with Snowflake configuration."""
    repo_root = get_repo_root()

    # Determine the test path
    if not test_path:
        test_path = "tests/end2end"
    elif test_path == "early_access":
        test_path = "tests/early_access/end2end"

    # Make sure the test_path is an absolute path
    if not os.path.isabs(test_path):
        test_path = os.path.join(repo_root, test_path)

    # Build the command
    cmd = ["pytest", test_path]

    # Add options
    if update:
        cmd.append("--snapshot-update")
    if verbose:
        cmd.append("-v")

    divider()
    rich.print("[yellow bold]Running end-to-end tests")
    rich.print(f"[yellow]Test path: {test_path}")
    rich.print("")

    # Run the command
    result = subprocess.run(cmd)

    divider()
    sys.exit(result.returncode)

@test_group.command("all")
@click.argument('test_path', required=False)
@click.option('--update', is_flag=True, help='Update snapshot files')
@click.option('--verbose', '-v', is_flag=True, help='Verbose output')
def run_all_tests(test_path, update, verbose):
    """Run both unit and end-to-end tests."""
    # First run unit tests
    unit_result = subprocess.run(
        ["rai-dev", "test", "unit"] +
        ([test_path] if test_path else []) +
        (["--update"] if update else []) +
        (["--verbose"] if verbose else [])
    ).returncode

    # Then run e2e tests
    e2e_result = subprocess.run(
        ["rai-dev", "test", "e2e"] +
        ([test_path] if test_path else []) +
        (["--update"] if update else []) +
        (["--verbose"] if verbose else [])
    ).returncode

    # Exit with non-zero if either test suite failed
    if unit_result != 0 or e2e_result != 0:
        sys.exit(1)
    return 0

#--------------------------------------------------
# Build Commands
#--------------------------------------------------

@cli.command("build")
def build_command():
    """Build the package."""
    result = subprocess.run(["python", "-m", "build"])
    sys.exit(result.returncode)

@cli.command("build:zip")
def build_zip_command():
    """Build the Snowflake notebook zip file."""
    result = subprocess.run(["python", "-m", "tools.snowflake_notebook.stage_zip"])
    sys.exit(result.returncode)

#--------------------------------------------------
# Type Checking
#--------------------------------------------------

@cli.command("typecheck")
@click.option('--path', default='', help='Path to typecheck')
def typecheck(path):
    """Run pyright type checking."""
    divider()
    rich.print("[yellow bold]Running type checking with pyright")
    rich.print("")

    result = subprocess.run(["pyright"] + ([path] if path else []))

    divider()
    sys.exit(result.returncode)

#--------------------------------------------------
# Linting
#--------------------------------------------------

@cli.command("lint")
@click.option('--path', default='', help='Path to lint')
def lint(path):
    """Run ruff linting."""
    divider()
    rich.print("[yellow bold]Running linting with ruff")
    rich.print("")

    result = subprocess.run(["ruff", "check"] + ([path] if path else []))

    divider()
    sys.exit(result.returncode)

#--------------------------------------------------
# Syncing Dependencies
#--------------------------------------------------

@cli.command("sync")
def sync():
    """Sync dependencies."""
    divider()
    rich.print("[yellow bold]Syncing dependencies")
    rich.print("")

    result = subprocess.run(["python", "-m", "pip", "install", "-r", "requirements.lock"])

    divider()
    sys.exit(result.returncode)

#--------------------------------------------------
# Update Lockfile
#--------------------------------------------------

@cli.command("lockfile", help="Update requirements.lock file with latest dependencies.")
@click.argument("update", required=True, default=None)
def update_lockfile(update):
    LOCKFILE = "requirements.lock"
    if update != "update":
        rich.print(f"[yellow]Do `rai-dev lockfile update` to update {LOCKFILE}")
        return
    """Update requirements lockfile with latest dependencies."""
    divider()
    rich.print(f"[yellow bold]Updating {LOCKFILE}")
    rich.print("")

    PYPROJECT_FILE = "pyproject.toml"
    REQUIREMENTS_LOCK = LOCKFILE

    # Ensure pip-tools is installed
    if shutil.which("pip-compile") is None:
        rich.print("pip-tools not found, installing it...")
        subprocess.run([sys.executable, "-m", "pip", "install", "pip-tools"], check=True)

    # Run pip-compile to generate requirements.lock
    rich.print(f"Running pip-compile to generate {REQUIREMENTS_LOCK}...")
    subprocess.run(
        [
            sys.executable,
            "-m",
            "piptools",
            "compile",
            "--quiet",
            "--no-strip-extras",
            "--upgrade",
            "--resolver=backtracking",
            "--extra",
            "dev",
            "--extra",
            "ea",
            "-o",
            REQUIREMENTS_LOCK,
            PYPROJECT_FILE,
        ],
        check=True
    )

    # Append -e . to the requirements file
    with open(REQUIREMENTS_LOCK, "a") as f:
        f.write("\n\n# Add the project root as a local package\n")
        f.write("-e .\n")

    rich.print(f"✔ Updated {REQUIREMENTS_LOCK}")

    divider()



# type: ignore
import dataclasses
from typing import Optional

from dataclasses_json import LetterCase, dataclass_json


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class ValueType:
    id: str
    name: str
    data_type: str
    data_type_name: str


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class EntityType:
    id: str
    name: str
    domain: list[str] = dataclasses.field(default_factory=list)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class UniqueConstraint:
    id: str
    roles: list[str]
    is_preferred_identifier: bool


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class MandatoryConstraint:
    id: str
    roles: list[str]


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class Role:
    id: str
    role_player: str
    role_player_name: str
    name: Optional[str] = None
    pre_bound_text: Optional[str] = None
    post_bound_text: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class Reading:
    id: str
    text: str
    roles: list[str]


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class DataType:
    id: str
    name: str


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class Constraints:
    uniqueness: list[UniqueConstraint]
    mandatory: list[MandatoryConstraint]


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class Relationship:
    id: str
    name: str
    roles: list[Role]
    readings: list[Reading] = dataclasses.field(default_factory=list)
    is_subtype: bool = False


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class Concepts:
    value_types: list[ValueType]
    entity_types: list[EntityType]

@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class SubtypeArrow:
    id: str
    start: str
    start_name: str
    end: str
    end_name: str

@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class Model:
    id: str
    name: str
    is_primary: bool
    concepts: Concepts
    relationships: list[Relationship]
    constraints: Constraints
    data_types: list[DataType]
    subtype_arrows: list[SubtypeArrow]


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclasses.dataclass
class ModelObject:
    model: Model



from dataclasses import dataclass
import textwrap
import time
from typing import Optional, cast
import relationalai as rai
from snowflake.snowpark import Row

from relationalai.clients.snowflake import Provider


@dataclass
class Task:
    name: str
    database: str
    schema: str
    procedure: str
    provider: Provider = cast(rai.clients.snowflake.Provider, rai.Provider())
    last_execution_time: Optional[str] = None
    logs_table: Optional[str] = None

    def create(
        self,
        args=[],
        warehouse="MAIN_WH",
        schedule_every_min=False,
        use_wrapper_sproc=False,
        wrapper_sproc_name=None,
        task_result_table_name=None,
    ):
        arg_string = ", ".join(
            f"'{arg}'" if isinstance(arg, str) else str(arg) for arg in args
        )
        if schedule_every_min:
            sql_query = textwrap.dedent(f"""
            CREATE OR REPLACE TASK {self.database}.{self.schema}.{self.name}
            WAREHOUSE = {warehouse}
            SCHEDULE = 'USING CRON * * * * * UTC'
            AS
            CALL {self.database}.{self.schema}.{self.procedure}({arg_string});
            """)
        else:
            sql_query = textwrap.dedent(f"""
            CREATE OR REPLACE TASK {self.database}.{self.schema}.{self.name}
            WAREHOUSE = {warehouse}
            AS
            CALL {self.database}.{self.schema}.{self.procedure}({arg_string});
            """)
        return self.provider.sql(sql_query)

    def exec_sync(self, log=print):
        log("Task created.")
        self.last_execution_time = self.get_current_timestamp()
        log("Executing task...")
        self.execute()
        log("Polling task...")
        return self.poll(self.last_execution_time, log=log)

    def execute(self):
        self.last_execution_time = self.get_current_timestamp()
        sql_query = f"EXECUTE TASK {self.database}.{self.schema}.{self.name};"
        return self.provider.sql(sql_query)

    def delete(self):
        sql_query = f"""
        DROP TASK {self.database}.{self.schema}.{self.name};
        """
        try:
            return self.provider.sql(sql_query)
        except Exception as e:
            if "does not exist" in str(e).lower():
                return None
            raise e

    def get_current_timestamp(self) -> str:
        sql_query = "SELECT CURRENT_TIMESTAMP() AS ts;"
        response = self.provider.sql(sql_query)
        return cast(str, response[0]["TS"])

    def check(self, scheduled_after=None):
        if scheduled_after is None:
            scheduled_after = self.last_execution_time
        sql_query = f"""
        SELECT *
        FROM TABLE({self.database}.INFORMATION_SCHEMA.TASK_HISTORY(
            SCHEDULED_TIME_RANGE_START => CAST(DATEADD('second', -1, TIMESTAMP '{scheduled_after}') AS TIMESTAMP_LTZ),
            TASK_NAME => '{self.name}'
        ));
        """
        return self.provider.sql(sql_query)

    def poll(self, scheduled_after=None, timeout=300, interval=5, log=print):
        if scheduled_after is None:
            scheduled_after = self.last_execution_time
        start_time = time.time()
        while time.time() - start_time < timeout:
            response = cast(list, self.check(scheduled_after))
            if response:
                log("Task scheduled...")
                latest = response[-1]
                if latest["COMPLETED_TIME"]:
                    log("Task completed")
                    return latest.as_dict()
            else:
                log("Task not found in history yet...")
            time.sleep(interval)
        log("Timeout reached. Task not found in history.")
        return None

    def get_server_logs(
        self, filename: str, table_name: Optional[str] = None, all=False, log=print
    ):
        "Save logs from Snowflake to a local SQLite database."
        if table_name is None:
            table_name = self.logs_table or "logs"
        if not self.logs_table:
            raise ValueError("Logs table not provided.")

        # if your table doesn't have a `timestamp` column, specify `all` as True
        where_clause = (
            f"where timestamp > '{self.last_execution_time}'"
            if not all or self.last_execution_time is None
            else ""
        )
        sql_query = f"select * from {self.database}.{self.schema}.{self.logs_table} {where_clause};"
        logs = [row.as_dict() for row in cast(list, self.provider.sql(sql_query))]
        log(f"{len(logs)} logs retrieved")
        if not logs:
            return
        from sqlite_utils import Database
        from sqlite_utils.db import Table

        db = Database(filename)
        cast(Table, db[table_name]).insert_all(logs)
        db.close()
        log(f"Logs saved from {table_name} in Snowflake to {filename} locally.")
        
    def wait_for_task_executions(
        self,
        min_executions,
        timeout=20 * 60,
        poll_interval=30,
        log=print,
    ):
        start_time = time.time()
        while time.time() - start_time < timeout:
            sql_query = f"""
                SELECT STATE, ERROR_CODE, ERROR_MESSAGE, QUERY_ID, NAME, DATABASE_NAME, SCHEMA_NAME, QUERY_TEXT
                FROM TABLE({self.database}.INFORMATION_SCHEMA.TASK_HISTORY(TASK_NAME => '{self.name}'));
            """
            try:
                rows = self.provider.sql(sql_query, format="list")
            except Exception as e:
                raise Exception(f"Error querying task history: {e}")

            state_counts = {}
            failure_details = []
            for row in rows:
                row = cast(Row, row)
                state = row.STATE
                state_counts[state] = state_counts.get(state, 0) + 1

                if state in ("FAILED", "FAILED_AND_AUTO_SUSPENDED", "SKIPPED", "CANCELLED"):
                    error_code = row.ERROR_CODE or ""
                    error_message = row.ERROR_MESSAGE or ""
                    query_id = row.QUERY_ID or ""
                    name = row.NAME or ""
                    db_name = row.DATABASE_NAME or ""
                    schema_name = row.SCHEMA_NAME or ""
                    query_text = row.QUERY_TEXT or ""
                    failure_details.append(
                        f"STATE: {state} | ERROR_CODE: {error_code} | ERROR_MESSAGE: {error_message} | QUERY_ID: {query_id} | "
                        f"TASK_NAME: {name} | DATABASE: {db_name} | SCHEMA: {schema_name} | QUERY_TEXT: {query_text}"
                    )

            if failure_details:
                log("Task Execution Failed Details:")
                for detail in failure_details:
                    log(detail)
                raise Exception("Task encountered a failure state. See error details above")

            succeeded_count = state_counts.get("SUCCEEDED", 0)
            if succeeded_count >= min_executions:
                log(f"Task reached {succeeded_count} successful executions.")
                return

            if state_counts.get("SCHEDULED", 0) > 0 or state_counts.get("EXECUTING", 0) > 0:
                log("Task is still running or scheduled. Polling again...")
                time.sleep(poll_interval)
                continue

            raise Exception(f"Task is in an unexpected state: {state_counts}")

        raise Exception(
            f"Timeout reached: task did not reach {min_executions} successful executions within {timeout/60} minutes"
        )


def format_task_history(row):
    return (
        f"""error message: {row["ERROR_MESSAGE"]}"""
        if row["ERROR_MESSAGE"]
        else f"""successfully completed at: {row["COMPLETED_TIME"]}"""
    )




from .. import dsl
from ..std.graphs import unwrap

# --------------------------------------------------
# Ego network
# --------------------------------------------------
def ego_network(graph, node, hops):
    a, b = dsl.create_vars(2)
    dsl.global_ns.graphlib_experimental.ego_network(graph, unwrap(node), hops, a, b)
    la = graph.compute._lookup(a)
    lb = graph.compute._lookup(b)
    return (la, lb)



from typing import Optional

from relationalai.early_access.dsl.core.types.concepts import Concept as CoreConcept
from relationalai.early_access.dsl.types import Type
from relationalai.early_access.dsl.core.types import Type as CoreType


class Concept(Type, CoreConcept):

    def __init__(self, model, name, extends: Optional[CoreType]=None):
        super().__init__(model, name)
        CoreConcept.__init__(self, name, extends)


from relationalai.early_access.dsl.core.types.standard import RowId, Integer
from relationalai.early_access.dsl.ontologies.models import Model
from relationalai.early_access.dsl.types.values import ValueType


class TabularSource:
    def __init__(self, name: str, schema: dict[str, ValueType], model: Model):
        self._name = name
        self._schema = schema

        self.basic_type_schema = {
            k: "Int64" if v.root_unconstrained_type() == Integer else "string"
            for k, v in schema.items()
        }

        for column_name, column_type in schema.items():
            setattr(self, column_name,
                    model.external_relation(f"{self._name}:{column_name}", RowId, column_type))



from relationalai.early_access.dsl.core.exprs import contextStack
from relationalai.early_access.dsl.core.exprs.scalar import ScalarExpr
from relationalai.early_access.dsl.core.logic import LogicFragment


class Aggregation(LogicFragment, ScalarExpr):

    # Each Schema object is a ContextManager
    def __enter__(self):
        contextStack.push(self)  # open a new context for this Schema
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        contextStack.pop()

    def __init__(self, method):
        LogicFragment.__init__(self)
        self._method = method
        self._aggregates = None
        self._schema = self

    def aggregates(self, var):
        self._aggregates = var

    def display(self):
        vars = [v.display() for v in self._scalars.values()]

        args = ", ".join(vars)
        body = self.rel_formula()
        return f"{self._method}[" + args + ":" + body + "]"

    def grounded(self): return False

    def pprint(self): return self.display()

    @staticmethod
    def max(var):
        return Aggregation._agg(var, "max")

    @staticmethod
    def min(var):
        return Aggregation._agg(var, "min")

    @staticmethod
    def sum(var):
        return Aggregation._agg(var, "sum")

    @staticmethod
    def _agg(var, method):
        agg = Aggregation(method)
        var == agg
        return agg


import os
import re
from nicegui import ui
import json

#--------------------------------------------------
# Terminal nodes
#--------------------------------------------------

TERMINAL_NODES = [
    "query"
]

#--------------------------------------------------
# debug.jsonl helpers
#--------------------------------------------------

class SpanNode:
    def __init__(self, id, type, parent_id, start_timestamp, attrs=None):
        self.id = id
        self.type = type
        self.parent_id = parent_id
        self.start_timestamp = start_timestamp
        self.end_timestamp = None
        self.elapsed = None
        self.start_attrs = attrs or {}
        self.end_attrs = {}
        self.children = []

    def add_end_data(self, end_timestamp, elapsed, end_attrs=None):
        self.end_timestamp = end_timestamp
        self.elapsed = elapsed
        if end_attrs:
            # Merge end_attrs into attrs
            self.end_attrs.update(end_attrs)

    def add_child(self, child):
        self.children.append(child)

    def __str__(self, level=0):
        indent = "  " * level
        result = f"{indent}{self.type} ({self.id}):\n"
        result += f"{indent}  elapsed: {self.elapsed:.6f}s\n"

        if self.start_attrs or self.end_attrs:
            result += f"{indent}  start attributes:\n"
            for key, value in self.start_attrs.items():
                # Truncate long values for display
                if isinstance(value, str) and len(value) > 100:
                    value = value[:97] + "..."
                result += f"{indent}    {key}: {value}\n"
            result += f"{indent}  end attributes:\n"
            for key, value in self.end_attrs.items():
                # Truncate long values for display
                if isinstance(value, str) and len(value) > 100:
                    value = value[:97] + "..."
                result += f"{indent}    {key}: {value}\n"

        if self.children:
            result += f"{indent}  children:\n"
            for child in self.children:
                result += child.__str__(level + 2)

        return result

def parse_jsonl_to_tree(jsonl_content):
    lines = jsonl_content.strip().split('\n')
    nodes_by_id = {}
    root_nodes = []

    for line in lines:
        data = json.loads(line)
        event_type = data["event"]

        if event_type == "span_start":
            span = data["span"]
            node = SpanNode(
                id=span["id"],
                type=span["type"],
                parent_id=span["parent_id"],
                start_timestamp=span["start_timestamp"],
                attrs=span.get("attrs", {})
            )
            nodes_by_id[node.id] = node

            # Link to parent if exists
            if node.parent_id and node.parent_id in nodes_by_id:
                nodes_by_id[node.parent_id].add_child(node)
            elif node.parent_id is None:
                root_nodes.append(node)

        elif event_type == "span_end":
            node_id = data["id"]
            if node_id in nodes_by_id:
                nodes_by_id[node_id].add_end_data(
                    end_timestamp=data["end_timestamp"],
                    elapsed=data.get("elapsed", 0),
                    end_attrs=data.get("end_attrs", {})
                )

    return root_nodes

#--------------------------------------------------
# UI
#--------------------------------------------------

last_mod_time = None
current_json_objects = []
active_ix = None
active_item = None

def set_item(ix):
    global active_item, active_ix
    if active_ix != ix:
        active_ix = ix
        active_item = current_json_objects[ix]
        sidebar.refresh()
        details.refresh()

def format_time(t):
    if not t:
        return "..."
    if t > 1:
        return f"{t:.1f}s"
    elif t > 0.001:
        return f"{t*1000:.1f}ms"
    elif t > 0.0005:
        return f"{t*1000:.2f}ms"
    else:
        return f"{t*1000000:.0f}us"

def header(text):
    return ui.label(text).style("font-size: 1.3em; font-weight: bold;")

def replace_long_brace_contents(code_str):
    def replacement(match):
        string = match.group(0)
        if len(string) > 300:
            # Extract the first 50 and last 30 characters from the string within brackets
            return '{' + string[1:51] + '...' + string[-31:-1] + '}'
        else:
            # If the string is not longer than 300 characters, return it unchanged
            return string

    # This regex matches sequences of characters wrapped in { }
    brace_content_regex = r'\{[\s\S]*?\}'

    # Use the sub method to replace the matched strings with the result of the replacement function
    return re.sub(brace_content_regex, replacement, code_str)


def code(c, language="python"):
    c = replace_long_brace_contents(c)
    c = re.sub(r"→", "->", c)
    c = re.sub(r"⇑", "^", c)
    return ui.code(c, language=language).style("padding-right: 30px").classes("w-full")

@ui.refreshable
def details():
    if active_item:
        if active_item['event'] == "compilation":
            header(f"{active_item['source']['file']}: {active_item['source']['line']}")
            # with ui.row():
            code(active_item["source"]["block"] or active_item["emitted"])
            code(f"{active_item['emitted']}")
            header("IR")
            code(f"{active_item['task']}")
            header("Rewritten")
            code(f"{active_item['passes'][-1]['task']}")
            header("Passes")
            for p in active_item['passes']:
                with ui.column().classes("w-full"):
                    with ui.row():
                        ui.label(p['name'])
                        ui.label(f"({p['elapsed']*1000000:.0f} us)")
                    code(p['task'])
        elif active_item['event'] == "time":
            ui.label(f"{active_item['type']} | {format_time(active_item['elapsed'])} | {active_item['results']['count']}")
            vals = active_item['results']['values']
            if len(vals):
                keys = [k for k in vals[0].keys()]
                columns = [{'name': k, 'label': k, 'field': k, "align": "left"} for k in keys]
                ui.table(columns=columns, rows=vals)
            if "code" in active_item:
                header("Code")
                code(active_item["code"])
            # ui.label(f"{active_item['results']['values']}").style('white-space: pre;')

def handle_attributes(attrs):
    if attrs.get("file"):
        ui.label(f"{attrs['file']}: {attrs['line']}")
    if attrs.get("source"):
        code(attrs["source"])
    if attrs.get("txn_id"):
        ui.label(f"txn_id: {attrs['txn_id']}")
    if attrs.get("name"):
        ui.label(f"{attrs['name']}")
    if attrs.get("code"):
        code(attrs["code"])
    if attrs.get("dsl"):
        code(attrs["dsl"])
    if attrs.get("metamodel"):
        code(attrs["metamodel"])
    if attrs.get("rel"):
        code(attrs["rel"])
    if attrs.get("results"):
        vals = attrs['results']
        if len(vals):
            keys = [k for k in vals[0].keys()]
            columns = [{'name': k, 'label': k, 'field': k, "align": "left"} for k in keys]
            ui.table(columns=columns, rows=vals)

def handle_body(span: SpanNode):
    handle_attributes(span.start_attrs)
    if span.children:
        for child in span.children:
            span_ui(child)
    handle_attributes(span.end_attrs)

def span_ui(span: SpanNode):
    with ui.column().style("background:#33333399; padding:5px 10px; margin:5px; "):
        if span.children or span.start_attrs or span.end_attrs:
            with ui.expansion(f"{span.type}",
                            caption=f"{format_time(span.elapsed)}",
                            value=True).classes('w-full').style("padding:0; margin:0;"):
                handle_body(span)
        else:
            with ui.row().style("padding:0; margin:0;"):
                ui.label(f"{span.type}")
                ui.label(f"{format_time(span.elapsed)}").style("color:#999; margin:0;")
                handle_body(span)




@ui.refreshable
def sidebar():
    with ui.column():
        for root in current_json_objects:
            span_ui(root)



def poll():
    global last_mod_time, active_item
    global current_json_objects
    # Check the last modification time of the file
    try:
        mod_time = os.path.getmtime('debug.jsonl')
        if last_mod_time is None or mod_time > last_mod_time:
            last_mod_time = mod_time
            # File has changed, read and parse the new content
            with open('debug.jsonl', 'r') as file:
                content = file.read()
                if content:
                    new_tree = parse_jsonl_to_tree(content)
                    # Update the current JSON objects
                    current_json_objects = new_tree

                    if active_ix is not None and len(current_json_objects) > active_ix:
                        active_item = current_json_objects[active_ix]
                    # Refresh the UI with the new objects
                    sidebar.refresh()
                    details.refresh()
    except FileNotFoundError:
        pass

def main(host="0.0.0.0", port=8080):
    ui.dark_mode().enable()
    with ui.row():
        with ui.column() as c:
            c.style("cursor: pointer;")
            sidebar()
        with ui.column() as c:
            c.style("padding-left: 2em;")
            details()

    ui.timer(1, poll)
    ui.run(reload=False, host=host, port=port)

if __name__ in {"__main__", "__mp_main__"}:
    main()



from __future__ import annotations
from dataclasses import dataclass, field
import time
from typing import Any, Optional, Tuple, TypeVar, Union

from relationalai import debugging
from relationalai.early_access.metamodel import ir
from relationalai.early_access.metamodel.observer import Observer
from relationalai.early_access.metamodel.util import OrderedSet, ordered_set, FrozenOrderedSet



@dataclass(frozen=True)
class Compiler():
    """ Compilers can rewrite a model into a different model, and can compile a model into
    a String, usually of a different language like Rel or SQL. """
    # configurable sequence of passes
    passes: list[Pass]

    def rewrite(self, model: ir.Model, observer: Optional[Observer]=None, options:dict={}) -> ir.Model:
        """ Apply a sequence of transformation passes over a model, creating a new model. """

        with debugging.span("passes") as span:
            for p in self.passes:
                start = time.perf_counter()
                with debugging.span(p.name) as span:
                    model = p.walk(model)
                    span["metamodel"] = str(model.root)
                if observer is not None:
                    observer.update(p.name, model, time.perf_counter() - start)
                p.reset()
        return model

    def compile(self, model: ir.Model, observer: Optional[Observer]=None, options:dict={}) -> str:
        model = self.rewrite(model, observer, options)
        start = time.perf_counter()
        result = self.do_compile(model, options)
        if observer is not None:
            observer.update("compile", result, time.perf_counter() - start)
        return result

    def do_compile(self, model: ir.Model, options:dict={}) -> str:
        """ Perform the compilation from model to string. The default implementation simply
        pretty prints the model IR. """
        return ir.node_to_string(model)

@dataclass
class Pass():
    """ A traversal over the model.

    This class provides handler methods for all types of nodes in the IR.

    The default implementation assumes that we will start traversal with a `Model` node, and
    that we are only interested in rewriting the model's root task. Therefore, the default
    handler implementation for engines and the data model simply return the node being
    traversed, as we don't want to modify them by default.

    Sub-classes can overwrite any of the handler methods to provide custom modifications.
    """
    # TODO: do we really need this?
    seen:OrderedSet[Any]  = field(default_factory=ordered_set)

    @property
    def name(self):
        return self.__class__.__name__

    def reset(self):
        self.seen.clear()

    T = TypeVar('T', bound=Union[ir.Node, ir.Value])
    def walk(self, node: T, parent=None, ctx:Optional[Any]=None) -> T:
        # if node is a value, just return it
        if not isinstance(node, ir.Node):
            return node

        # node is actually some Node type, handle with the appropriate handler
        handler = getattr(self, f"handle_{node.kind}", None)
        if handler:
            return handler(node, parent, ctx)
        else:
            raise NotImplementedError(f"walk: {node.kind}")

    def walk_set(self, items: FrozenOrderedSet[T], parent=None, ctx:Optional[Any]=None) -> FrozenOrderedSet[T]:
        return ordered_set(*[self.walk(n, parent, ctx) for n in items]).frozen()

    def walk_list(self, items: Tuple[T, ...], parent=None, ctx:Optional[Any]=None) -> Tuple[T, ...]:
        return tuple([self.walk(n, parent, ctx) for n in items])

    U = TypeVar('U', bound=Union[ir.Node, ir.Value, FrozenOrderedSet[ir.Relation]])

    def walk_dict(self, items: dict[str, U], parent=None, ctx:Optional[Any]=None) -> dict[str, U]:
        d = dict()
        for k, v in items.items():
            if isinstance(v, FrozenOrderedSet):
                d[k] = self.walk_set(v, parent, ctx) # type: ignore
            else:
                d[k] = self.walk(v, parent, ctx)
        return d



    #-------------------------------------------------
    # Public Types - Model
    #-------------------------------------------------

    def handle_model(self, model: ir.Model, parent: None, ctx:Optional[Any]=None):
        return ir.Model(
            self.walk_set(model.engines, model),
            self.walk_set(model.relations, model),
            self.walk_set(model.types, model),
            self.walk(model.root, model),
        )

    #-------------------------------------------------
    # Public Types - Engine
    #-------------------------------------------------

    def handle_capability(self, node: ir.Capability, parent: ir.Node, ctx:Optional[Any]=None):
        return node

    def handle_engine(self, node: ir.Engine, parent: ir.Node, ctx:Optional[Any]=None):
        return node

    #-------------------------------------------------
    # Public Types - Data Model
    #-------------------------------------------------

    def handle_scalartype(self, node: ir.ScalarType, parent: ir.Node, ctx:Optional[Any]=None):
        return node

    def handle_listtype(self, node: ir.ListType, parent: ir.Node, ctx:Optional[Any]=None):
        return node

    def handle_settype(self, node: ir.SetType, parent: ir.Node, ctx:Optional[Any]=None):
        return node

    def handle_uniontype(self, node: ir.UnionType, parent: ir.Node, ctx:Optional[Any]=None):
        # TODO - we could traverse the children
        return node

    def handle_field(self, node: ir.Field, parent: ir.Node, ctx:Optional[Any]=None):
        return node

    def handle_relation(self, node: ir.Relation, parent: ir.Node, ctx:Optional[Any]=None):
        return node

    #-------------------------------------------------
    # Public Types - Tasks
    #-------------------------------------------------

    def handle_task(self, node: ir.Task, parent: ir.Node, ctx:Optional[Any]=None):
        return node

    #
    # Task composition
    #

    def handle_logical(self, node: ir.Logical, parent: ir.Node, ctx:Optional[Any]=None):
        return ir.Logical(
            node.engine,
            self.walk_list(node.hoisted, node, ctx),
            self.walk_list(node.body, node, ctx)
        )

    def handle_union(self, node: ir.Union, parent: ir.Node, ctx:Optional[Any]=None):
        return ir.Union(
            node.engine,
            self.walk_list(node.hoisted, node, ctx),
            self.walk_list(node.tasks, node, ctx)
        )

    def handle_sequence(self, node: ir.Sequence, parent: ir.Node, ctx:Optional[Any]=None):
        return ir.Sequence(
            node.engine,
            self.walk_list(node.hoisted, node, ctx),
            self.walk_list(node.tasks, node, ctx)
        )

    def handle_match(self, node: ir.Match, parent: ir.Node, ctx:Optional[Any]=None):
        return ir.Match(
            node.engine,
            self.walk_list(node.hoisted, node, ctx),
            self.walk_list(node.tasks, node, ctx)
        )

    def handle_until(self, node: ir.Until, parent: ir.Node, ctx:Optional[Any]=None):
        return ir.Until(
            node.engine,
            self.walk_list(node.hoisted, node, ctx),
            self.walk(node.check, node, ctx),
            self.walk(node.body, node, ctx)
        )

    def handle_wait(self, node: ir.Wait, parent: ir.Node, ctx:Optional[Any]=None):
        return ir.Wait(
            node.engine,
            self.walk_list(node.hoisted, node, ctx),
            self.walk(node.check, node, ctx)
        )

    #
    # Relational Operations
    #

    def handle_var(self, node: ir.Var, parent: ir.Node, ctx:Optional[Any]=None):
        return ir.Var(
            self.walk(node.type, node, ctx),
            node.name or f"v{node.id}"
        )

    def handle_default(self, node: ir.Default, parent: ir.Node, ctx:Optional[Any]=None):
        return ir.Default(
            self.walk(node.var, node, ctx),
            node.value
        )

    def handle_literal(self, node: ir.Literal, parent: ir.Node, ctx:Optional[Any]=None):
        return ir.Literal(
            self.walk(node.type, node, ctx),
            node.value
        )

    def handle_update(self, node: ir.Update, parent: ir.Node, ctx:Optional[Any]=None):
        return ir.Update(
            node.engine,
            self.walk(node.relation, node, ctx),
            self.walk_list(node.args, node, ctx),
            node.effect,
            node.annotations
        )

    def handle_lookup(self, node: ir.Lookup, parent: ir.Node, ctx:Optional[Any]=None):
        return ir.Lookup(
            node.engine,
            self.walk(node.relation, node, ctx),
            self.walk_list(node.args, node, ctx)
        )

    def handle_output(self, node: ir.Output, parent: ir.Node, ctx:Optional[Any]=None):
        s = ordered_set()
        for k, v in node.aliases:
            s.add((k, self.walk(v, ctx)))
        return ir.Output(node.engine, s.frozen())

    def handle_construct(self, node: ir.Construct, parent: ir.Node, ctx:Optional[Any]=None):
        return ir.Construct(
            node.engine,
            self.walk_list(node.values, node, ctx),
            node.id_var
        )

    def handle_aggregate(self, node: ir.Aggregate, parent: ir.Node, ctx:Optional[Any]=None):
        return ir.Aggregate(
            node.engine,
            self.walk(node.aggregation, node, ctx),
            self.walk_list(node.projection, node, ctx),
            self.walk_list(node.group, node, ctx),
            self.walk_list(node.args, node, ctx)
        )

    #
    # Logical Quantifiers
    #

    def handle_not(self, node: ir.Not, parent: ir.Node, ctx:Optional[Any]=None):
        return ir.Not(
            node.engine,
            self.walk(node.task, node, ctx)
        )

    def handle_exists(self, node: ir.Exists, parent: ir.Node, ctx:Optional[Any]=None):
        return ir.Exists(
            node.engine,
            self.walk_list(node.vars, node, ctx),
            self.walk(node.task, node, ctx)
        )

    def handle_forall(self, node: ir.ForAll, parent: ir.Node, ctx:Optional[Any]=None):
        return ir.ForAll(
            node.engine,
            self.walk_list(node.vars, node, ctx),
            self.walk(node.task, node, ctx),
        )

    #
    # Iteration (Loops)
    #
    def handle_loop(self, node: ir.Loop, parent: ir.Node, ctx:Optional[Any]=None):
        return ir.Loop(
            node.engine,
            self.walk_list(node.hoisted, node, ctx),
            self.walk(node.iter, node, ctx),
            self.walk(node.body, node, ctx),
        )

    def handle_break(self, node: ir.Break, parent: ir.Node, ctx:Optional[Any]=None):
        return ir.Break(
            node.engine,
            self.walk(node.check, node, ctx),
        )


class Clone(Pass):
    """ A pass that does not, thereby cloning the IR. """
    pass
